/*
 * libmad - MPEG audio decoder library
 * Copyright (C) 2000-2001 Robert Leslie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: synth.c,v 1.21 2001/10/18 08:11:43 rob Exp $
 */

#include "audio.h"

static struct audio_dither left_dither, right_dither;
static struct audio_stats audio;

/*
 * NAME:	clip()
 * DESCRIPTION:	gather signal statistics while clipping
 */
static void clip(mad_fixed_t *sample)
{
  	enum {
    	MIN = -MAD_F_ONE,
    	MAX =  MAD_F_ONE - 1
  	};

  	if (*sample >= audio.peak_sample)
  	{
    	if (*sample > MAX)
    	{
      		++audio.clipped_samples;
      		if (*sample - MAX > audio.peak_clipping)
				audio.peak_clipping = *sample - MAX;

      		*sample = MAX;
    	}
    	audio.peak_sample = *sample;
  	}
  	else if (*sample < -audio.peak_sample)
  	{
    	if (*sample < MIN)
    	{
      		++audio.clipped_samples;
      		if (MIN - *sample > audio.peak_clipping)
				audio.peak_clipping = MIN - *sample;

      		*sample = MIN;
    	}
    	audio.peak_sample = -*sample;
  	}
}

/*
 * NAME:	audio_linear_round()
 * DESCRIPTION:	generic linear sample quantize routine
 */
signed long audio_linear_round(unsigned int bits, mad_fixed_t sample)
{
  	/* round */
  	sample += (1L << (MAD_F_FRACBITS - bits));

  	/* clip */
  	clip(&sample);

  	/* quantize and scale */
  	return sample >> (MAD_F_FRACBITS + 1 - bits);
}

/*
 * NAME:	prng()
 * DESCRIPTION:	32-bit pseudo-random number generator
 */
static unsigned long prng(unsigned long state)
{
  	return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

/*
 * NAME:	audio_linear_dither()
 * DESCRIPTION:	generic linear sample quantize and dither routine
 */
static signed long audio_linear_dither(unsigned int bits, mad_fixed_t sample, struct audio_dither *dither)
{
  	unsigned int scalebits;
  	mad_fixed_t output, mask, random;

  	enum {
    	MIN = -MAD_F_ONE,
    	MAX =  MAD_F_ONE - 1
  	};

  	/* noise shape */
  	sample += dither->error[0] - dither->error[1] + dither->error[2];

  	dither->error[2] = dither->error[1];
  	dither->error[1] = dither->error[0] / 2;

  	/* bias */
  	output = sample + (1L << (MAD_F_FRACBITS + 1 - bits - 1));

  	scalebits = MAD_F_FRACBITS + 1 - bits;
  	mask = (1L << scalebits) - 1;

  	/* dither */
  	random  = prng(dither->random);
  	output += (random & mask) - (dither->random & mask);

  	dither->random = random;

  	/* clip */
  	if (output >= audio.peak_sample)
  	{
    	if (output > MAX)
    	{
      		++audio.clipped_samples;
      		if (output - MAX > audio.peak_clipping)
				audio.peak_clipping = output - MAX;

      		output = MAX;

      		if (sample > MAX)
				sample = MAX;
    	}
    	audio.peak_sample = output;
  	}
  	else if (output < -audio.peak_sample)
 	{
    	if (output < MIN)
    	{
      		++audio.clipped_samples;
      		if (MIN - output > audio.peak_clipping)
				audio.peak_clipping = MIN - output;

      		output = MIN;

      		if (sample < MIN)
				sample = MIN;
    	}
    	audio.peak_sample = -output;
  	}

  	/* quantize */
  	output &= ~mask;

  	/* error feedback */
  	dither->error[0] = sample - output;

  	/* scale */
  	return output >> scalebits;
}

/*
 * NAME:	audio_pcm_s16le()
 * DESCRIPTION:	write a block of signed 16-bit little-endian PCM samples
 */
unsigned int audio_pcm_s16le(unsigned char *data, unsigned int nsamples, mad_fixed_t const *left, mad_fixed_t const *right)
{
  	int nCount;
  	register signed int sample0, sample1;

    for (nCount = 0; nCount < (int) nsamples; nCount ++)
	{
		//sample0 = audio_linear_dither(16, *left++,  &left_dither);
		//sample1 = audio_linear_dither(16, *right++, &right_dither);

		sample0 = audio_linear_round(16, *left++);
		sample1 = audio_linear_round(16, *right++);

		data[0] = sample0 >> 0;
		data[1] = sample0 >> 8;
		data[2] = sample1 >> 0;
		data[3] = sample1 >> 8;

		data += 4;
	}

    return nsamples * 2 * 2;
}

/*
 * NAME:	audio_pcm_s16be()
 * DESCRIPTION:	write a block of signed 16-bit big-endian PCM samples
 */
unsigned int audio_pcm_s16be(unsigned char *data, unsigned int nsamples, mad_fixed_t const *left, mad_fixed_t const *right)
{
  	int nCount;
  	register signed int sample0, sample1;

    for (nCount = 0; nCount < (int) nsamples; nCount ++)
	{
		//sample0 = audio_linear_dither(16, *left++,  &left_dither);
		//sample1 = audio_linear_dither(16, *right++, &right_dither);

		sample0 = audio_linear_round(16, *left++);
		sample1 = audio_linear_round(16, *right++);

		data[0] = sample0 >> 8;
		data[1] = sample0 >> 0;
		data[2] = sample1 >> 8;
		data[3] = sample1 >> 0;

		data += 4;
	}

    return nsamples * 2 * 2;
}

void audio_init(void)
{
  	audio.clipped_samples = 0;
  	audio.peak_clipping   = 0;
  	audio.peak_sample     = 0;
}