/*
 * libmad - MPEG audio decoder library
 * Copyright (C) 2000-2001 Robert Leslie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: fixed.h,v 1.30 2001/11/02 09:51:06 rob Exp $
 */

#ifndef LIBMAD_FIXED_H
#define LIBMAD_FIXED_H

#define FPM_DEFAULT
//#define OPT_SPEED

#define ASO_INTERLEAVE2

typedef   signed int mad_fixed_t;

typedef   signed int mad_fixed64hi_t;
typedef unsigned int mad_fixed64lo_t;

/*
 * Fixed-point format: 0xABBBBBBB
 * A == whole part      (sign + 3 bits)
 * B == fractional part (28 bits)
 *
 * Values are signed two's complement, so the effective range is:
 * 0x80000000 to 0x7fffffff
 *       -8.0 to +7.9999999962747097015380859375
 *
 * The smallest representable value is:
 * 0x00000001 == 0.0000000037252902984619140625 (i.e. about 3.725e-9)
 *
 * 28 bits of fractional accuracy represent about
 * 8.6 digits of decimal accuracy.
 *
 * Fixed-point numbers can be added or subtracted as normal
 * integers, but multiplication requires shifting the 64-bit result
 * from 56 fractional bits back to 28 (and rounding.)
 *
 * Changing the definition of MAD_F_FRACBITS is only partially
 * supported, and must be done with care.
 */

#define MAD_F_FRACBITS		28

#define MAD_F(x)			((mad_fixed_t) (x##L))

#define MAD_F_MIN			((mad_fixed_t) -0x80000000L)
#define MAD_F_MAX			((mad_fixed_t) +0x7fffffffL)

#define MAD_F_ONE			MAD_F(0x10000000)

#define mad_f_tofixed(x)	((mad_fixed_t)  \
							((x) * (double) (1L << MAD_F_FRACBITS) + 0.5))
#define mad_f_todouble(x)	((double)  \
							((x) / (double) (1L << MAD_F_FRACBITS)))

#define mad_f_intpart(x)	((x) >> MAD_F_FRACBITS)
#define mad_f_fracpart(x)	((x) & ((1L << MAD_F_FRACBITS) - 1))
				/* (x should be positive) */

#define mad_f_fromint(x)	((x) << MAD_F_FRACBITS)

#define mad_f_add(x, y)		((x) + (y))
#define mad_f_sub(x, y)		((x) - (y))

/* --- Intel --------------------------------------------------------------- */
#if defined(FPM_INTEL)

#if defined(_MSC_VER)
#pragma warning(push)
#pragma warning(disable: 4035)  /* no return value */
static __forceinline
mad_fixed_t mad_f_mul_inline(mad_fixed_t x, mad_fixed_t y)
{
	enum {
		fracbits = MAD_F_FRACBITS
	};

	__asm {
		mov eax, x
		imul y
		shrd eax, edx, fracbits
	}

	/* implicit return of eax */
}
#pragma warning(pop)

#define mad_f_mul		mad_f_mul_inline
#define mad_f_scale64
#else
/*
 * This Intel version is fast and accurate; the disposition of the least
 * significant bit depends on OPT_ACCURACY via mad_f_scale64().
 */
#define MAD_F_MLX(hi, lo, x, y)  \
    asm ("imull %3"  \
	 : "=a" (lo), "=d" (hi)  \
	 : "%a" (x), "rm" (y)  \
	 : "cc")

#if defined(OPT_ACCURACY)
/*
 * This gives best accuracy but is not very fast.
 */
#define MAD_F_MLA(hi, lo, x, y)  \
    ({ mad_fixed64hi_t __hi;  \
       mad_fixed64lo_t __lo;  \
       MAD_F_MLX(__hi, __lo, (x), (y));  \
       asm ("addl %2,%0\n\t"  \
	    "adcl %3,%1"  \
	    : "=rm" (lo), "=rm" (hi)  \
	    : "r" (__lo), "r" (__hi), "0" (lo), "1" (hi)  \
	    : "cc");  \
    })
#endif  /* OPT_ACCURACY */

#if defined(OPT_ACCURACY)
/*
 * Surprisingly, this is faster than SHRD followed by ADC.
 */
#define mad_f_scale64(hi, lo)  \
    ({ mad_fixed64hi_t __hi_;  \
       mad_fixed64lo_t __lo_;  \
       mad_fixed_t __result;  \
       asm ("addl %4,%2\n\t"  \
	    "adcl %5,%3"  \
	    : "=rm" (__lo_), "=rm" (__hi_)  \
	    : "0" (lo), "1" (hi),  \
	      "ir" (1L << (MAD_F_SCALEBITS - 1)), "ir" (0)  \
	    : "cc");  \
       asm ("shrdl %3,%2,%1"  \
	    : "=rm" (__result)  \
	    : "0" (__lo_), "r" (__hi_), "I" (MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
#else
#define mad_f_scale64(hi, lo)  \
    ({ mad_fixed_t __result;  \
       asm ("shrdl %3,%2,%1"  \
	    : "=rm" (__result)  \
	    : "0" (lo), "r" (hi), "I" (MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
#endif  /* OPT_ACCURACY */

#define MAD_F_SCALEBITS  MAD_F_FRACBITS
#endif

/* --- ARM ----------------------------------------------------------------- */
#elif defined(FPM_ARM)

/*
 * This ARM V4 version is as accurate as FPM_64BIT but much faster. The
 * least significant bit is properly rounded at no CPU cycle cost!
 */
#if 1
/*
 * There's a bug somewhere, possibly in the compiler, that sometimes makes
 * this necessary instead of the default implementation via MAD_F_MLX and
 * mad_f_scale64. It may be related to the use (or lack) of
 * -finline-functions and/or -fstrength-reduce.
 *
 * This is also apparently faster than MAD_F_MLX/mad_f_scale64.
 */
#define mad_f_mul(x, y)  \
    ({ mad_fixed64hi_t __hi;  \
       mad_fixed64lo_t __lo;  \
       mad_fixed_t __result;  \
       asm ("smull	%0, %1, %3, %4\n\t"  \
	    "movs	%0, %0, lsr %5\n\t"  \
	    "adc	%2, %0, %1, lsl %6"  \
	    : "=&r" (__lo), "=&r" (__hi), "=r" (__result)  \
	    : "%r" (x), "r" (y),  \
	      "M" (MAD_F_SCALEBITS), "M" (32 - MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
#endif

#define MAD_F_MLX(hi, lo, x, y)  \
    asm ("smull	%0, %1, %2, %3"  \
	 : "=&r" (lo), "=&r" (hi)  \
	 : "%r" (x), "r" (y))

#define MAD_F_MLA(hi, lo, x, y)  \
    asm ("smlal	%0, %1, %2, %3"  \
	 : "+r" (lo), "+r" (hi)  \
	 : "%r" (x), "r" (y))

#define MAD_F_MLN(hi, lo)  \
    asm ("rsbs	%0, %2, #0\n\t"  \
	 "rsc	%1, %3, #0"  \
	 : "=r" (lo), "=r" (hi)  \
	 : "0" (lo), "1" (hi)  \
	 : "cc")

#define mad_f_scale64(hi, lo)  \
    ({ mad_fixed_t __result;  \
       asm ("movs	%0, %1, lsr %3\n\t"  \
	    "adc	%0, %0, %2, lsl %4"  \
	    : "=r" (__result)  \
	    : "r" (lo), "r" (hi),  \
	      "M" (MAD_F_SCALEBITS), "M" (32 - MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })

#define MAD_F_SCALEBITS  MAD_F_FRACBITS

/* --- Default ------------------------------------------------------------- */
#elif defined(FPM_DEFAULT)

/*
 * This version is the most portable but it loses significant accuracy.
 * Furthermore, accuracy is biased against the second argument, so care
 * should be taken when ordering operands.
 *
 * The scale factors are constant as this is not used with SSO.
 *
 * Pre-rounding is required to stay within the limits of compliance.
 */
#if defined(OPT_SPEED)
#define mad_f_mul(x, y)	(((x) >> 12) * ((y) >> 16))
#else
#define mad_f_mul(x, y)	((((x) + (1L << 11)) >> 12) *  \
				 (((y) + (1L << 15)) >> 16))
#endif

/* ------------------------------------------------------------------------- */
#else
#error "no FPM selected"
#endif

/* default implementations */

#if !defined(mad_f_mul)
#define mad_f_mul(x, y)  \
		({ mad_fixed64hi_t __hi;  \
		   mad_fixed64lo_t __lo;  \
		   MAD_F_MLX(__hi, __lo, (x), (y));  \
		   mad_f_scale64(__hi, __lo);  \
		})
#endif

#if !defined(MAD_F_MLA)
#define MAD_F_ML0(hi, lo, x, y)	((lo)  = mad_f_mul((x), (y)))
#define MAD_F_MLA(hi, lo, x, y)	((lo) += mad_f_mul((x), (y)))
#define MAD_F_MLN(hi, lo)		((lo)  = -(lo))
#define MAD_F_MLZ(hi, lo)		((void) (hi), (mad_fixed_t) (lo))
#endif

#if !defined(MAD_F_ML0)
#define MAD_F_ML0(hi, lo, x, y)	MAD_F_MLX((hi), (lo), (x), (y))
#endif

#if !defined(MAD_F_MLN)
#define MAD_F_MLN(hi, lo)		((hi) = ((lo) = -(lo)) ? ~(hi) : -(hi))
#endif

#if !defined(MAD_F_MLZ)
#define MAD_F_MLZ(hi, lo)		mad_f_scale64((hi), (lo))
#endif

#if !defined(mad_f_scale64)
#if defined(OPT_ACCURACY)
#define mad_f_scale64(hi, lo)  \
		((((mad_fixed_t)  \
		(((hi) << (32 - (MAD_F_SCALEBITS - 1))) |  \
		((lo) >> (MAD_F_SCALEBITS - 1)))) + 1) >> 1)
#else
#define mad_f_scale64(hi, lo)  \
		((mad_fixed_t)  \
		(((hi) << (32 - MAD_F_SCALEBITS)) |  \
		((lo) >> MAD_F_SCALEBITS)))
#endif
#define MAD_F_SCALEBITS  MAD_F_FRACBITS
#endif

/* miscellaneous C routines */

mad_fixed_t mad_f_abs(mad_fixed_t);

#endif
